from __future__ import annotations

import json
import os
import time
from collections.abc import Iterable
from contextlib import suppress
from dataclasses import dataclass
from pathlib import Path

import httpx
from packaging.requirements import Requirement
from packaging.version import Version
from packaging.version import parse as parse_version


@dataclass(frozen=True)
class Outdated:
    """Represents a pinned requirement that is behind PyPI latest."""

    name: str
    current: str
    latest: str


@dataclass(frozen=True)
class Unpinned:
    """Represents a requirement that isn't pinned with '==' (we don't check it)."""

    name: str
    spec: str


class PyPICache:
    """
    Very small file cache for /pypi/{name}/json with ETag + TTL.
    File content: {"version": "...", "etag": "...", "ts": <epoch>}
    """

    def __init__(self, ttl_seconds: int = 86400, cache_dir: Path | None = None) -> None:
        base = Path(os.environ.get("XDG_CACHE_HOME", Path.home() / ".cache")).expanduser()
        self.dir = (cache_dir or (base / "animadao" / "pypi")).resolve()
        self.dir.mkdir(parents=True, exist_ok=True)
        self.ttl = ttl_seconds

    def _path(self, name: str) -> Path:
        return self.dir / f"{name.lower()}.json"

    def load(self, name: str) -> tuple[str | None, str | None, float]:
        p = self._path(name)
        if not p.is_file():
            return None, None, 0.0
        try:
            data = json.loads(p.read_text(encoding="utf-8"))
            return data.get("version"), data.get("etag"), float(data.get("ts", 0))
        except Exception:
            return None, None, 0.0

    def save(self, name: str, version: str, etag: str | None) -> None:
        p = self._path(name)
        payload = {"version": version, "etag": etag, "ts": time.time()}
        p.write_text(json.dumps(payload), encoding="utf-8")


class VersionChecker:
    """
    Backward-compatible version checker.

    - __init__ now accepts optional `requirements` (legacy behavior),
      plus keyword-only TTL & concurrency (reserved).
    - get_latest_version(name) is back (sync) for monkeypatching in tests.
    - check_declared(Optional[list[Requirement]]) and check_installed(mapping)
      compare against PyPI latest using cache.
    """

    PYPI_JSON = "https://pypi.org/pypi/{name}/json"

    def __init__(
        self,
        requirements: list[Requirement] | None = None,
        *,
        ttl_seconds: int = 86400,
        concurrency: int = 8,
    ) -> None:
        self._requirements: list[Requirement] = requirements or []
        self.cache = PyPICache(ttl_seconds=ttl_seconds)
        # `concurrency` зарезервирован под будущий async-bulk, здесь не критичен
        self.concurrency = max(1, int(concurrency))

    # -------- compatibility method (used by tests to monkeypatch) --------
    def get_latest_version(self, name: str) -> Version | None:
        """
        Return latest version from PyPI for `name`, with ETag/TTL cache.
        Sync on purpose so tests can monkeypatch it easily.
        """
        cached_ver, etag, ts = self.cache.load(name)
        # valid TTL?
        if cached_ver and (time.time() - ts) < self.cache.ttl:
            try:
                return parse_version(cached_ver)
            except Exception:
                pass

        headers = {}
        if etag:
            headers["If-None-Match"] = etag
        try:
            with httpx.Client(timeout=10.0) as client:
                r = client.get(self.PYPI_JSON.format(name=name), headers=headers)
                if r.status_code == 304 and cached_ver:
                    return parse_version(cached_ver)
                r.raise_for_status()
                data = r.json()
                v_str = data["info"]["version"]
                new_etag = r.headers.get("ETag")
                self.cache.save(name, v_str, new_etag)
                return parse_version(v_str)
        except Exception:
            if cached_ver:
                try:
                    return parse_version(cached_ver)
                except Exception:
                    return None
            return None

    # -------- declared --------
    def check_declared(
        self, requirements: Iterable[Requirement] | None = None
    ) -> tuple[list[Outdated], list[Unpinned]]:
        reqs = list(requirements) if requirements is not None else list(self._requirements)
        outdated: list[Outdated] = []
        unpinned: list[Unpinned] = []

        pins: dict[str, Version] = {}
        for req in reqs:
            try:
                if req.marker and not req.marker.evaluate():
                    continue
            except Exception:
                pass
            equals = [sp for sp in req.specifier if sp.operator == "=="]
            if not equals:
                spec = str(req.specifier) if str(req.specifier) else "*"
                unpinned.append(Unpinned(name=req.name, spec=spec))
            else:
                with suppress(Exception):
                    pins[req.name] = parse_version(equals[-1].version)

        use_monkeypatched = type(self).get_latest_version is not VersionChecker.get_latest_version

        for name, cur in pins.items():
            latest = self.get_latest_version(name) if use_monkeypatched else self.get_latest_version(name)
            if latest is not None and cur < latest:
                outdated.append(Outdated(name=name, current=str(cur), latest=str(latest)))

        return outdated, unpinned

    # -------- installed --------
    def check_installed(self, installed: dict[str, str]) -> tuple[list[Outdated], list[Unpinned]]:
        outdated: list[Outdated] = []
        for name, cur_str in installed.items():
            try:
                cur = parse_version(cur_str)
            except Exception:
                continue
            latest = self.get_latest_version(name)
            if latest is not None and cur < latest:
                outdated.append(Outdated(name=name, current=str(cur), latest=str(latest)))
        return outdated, []

    # -------- backward-compat shims --------
    def check(self):
        """
        Backward-compatible adapter for old tests:
        returns (outdated, unpinned) for declared requirements attached to the instance.
        """
        return self.check_declared()

    def check_versions(self):
        """
        Older name used previously in the project/tests.
        Delegates to `check_declared()`.
        """
        return self.check_declared()



from __future__ import annotations

import json
import sys
from collections.abc import Iterable
from pathlib import Path

import click

from animadao.config import load_config
from animadao.dependency_checker import guess_unused, load_declared_deps_any
from animadao.import_scanner import find_top_level_imports
from animadao.version_checker import VersionChecker


def _lower_set(items: Iterable[str] | None) -> set[str]:
    return {s.lower() for s in (items or [])}


@click.command(context_settings={"help_option_names": ["-h", "--help"]})
@click.option(
    "--project",
    type=click.Path(path_type=Path, exists=True, file_okay=False),
    default=Path("."),
    help="Project root.",
)
@click.option(
    "--mode",
    type=click.Choice(["declared", "installed"]),
    default=None,
    help="Check declared pins or installed packages.",
)
@click.option("--ignore", multiple=True, help="Ignore packages (case-insensitive). Can repeat.")
@click.option(
    "--pypi-ttl",
    type=int,
    default=None,
    help="PyPI cache TTL seconds (default from config or 86400).",
)
@click.option(
    "--pypi-concurrency",
    type=int,
    default=None,
    help="Parallel PyPI requests (default from config or 8).",
)
@click.option("--fail-if-outdated", is_flag=True, default=False, help="Fail if any outdated packages found.")
@click.option(
    "--fail-if-unpinned",
    is_flag=True,
    default=False,
    help="Fail if any unpinned requirements found (declared mode).",
)
@click.option(
    "--max-unused",
    type=int,
    default=None,
    help="Fail if count of unused declared deps exceeds this value (declared mode).",
)
def main(
    project: Path,
    mode: str | None,
    ignore: tuple[str, ...],
    pypi_ttl: int | None,
    pypi_concurrency: int | None,
    fail_if_outdated: bool,
    fail_if_unpinned: bool,
    max_unused: int | None,
) -> None:
    """
    Pre-commit gate for AnimaDao.
    Exits non-zero if selected policy checks fail.
    """
    cfg = load_config(project).with_overrides(mode=mode, ignore=ignore, ttl=pypi_ttl, conc=pypi_concurrency)
    ig = _lower_set(cfg.ignore_distributions)

    checker = VersionChecker(ttl_seconds=cfg.pypi_ttl_seconds, concurrency=cfg.pypi_concurrency)

    outdated = []
    unpinned = []
    unused: list[str] = []
    imports_found = 0
    declared_count = 0

    if cfg.mode == "declared":
        declared = load_declared_deps_any(project).requirements
        declared_count = len(declared)
        imports = find_top_level_imports(project)
        imports_found = len(imports)
        outdated, unpinned = checker.check_declared(declared)
        unused = guess_unused(declared, imports)
    else:
        from importlib import metadata as im

        installed = {d.metadata["Name"]: d.version for d in im.distributions()}
        outdated, _ = checker.check_installed(installed)

    # apply ignore
    outdated = [o for o in outdated if o.name.lower() not in ig]
    unpinned = [u for u in unpinned if u.name.lower() not in ig]
    unused = [u for u in unused if u.lower() not in ig]

    summary = {
        "mode": cfg.mode,
        "declared": declared_count,
        "imports_found": imports_found,
        "outdated": len(outdated),
        "unpinned": len(unpinned),
        "unused": len(unused),
    }

    print("AnimaDao summary:", json.dumps(summary, indent=2))

    violations: list[str] = []
    if fail_if_outdated and summary["outdated"] > 0:
        violations.append(f"outdated={summary['outdated']}")
    if cfg.mode == "declared":
        if fail_if_unpinned and summary["unpinned"] > 0:
            violations.append(f"unpinned={summary['unpinned']}")
        if max_unused is not None and summary["unused"] > max_unused:
            violations.append(f"unused={summary['unused']} > {max_unused}")

    if violations:
        print("AnimaDao policy violated:", ", ".join(violations))
        sys.exit(2)

    sys.exit(0)


if __name__ == "__main__":
    main()



from __future__ import annotations

import json
import sys
from collections.abc import Iterable
from pathlib import Path

import click
from packaging.requirements import Requirement

from animadao.config import load_config
from animadao.dependency_checker import guess_unused, load_declared_deps_any
from animadao.import_scanner import find_top_level_imports
from animadao.report_generator import generate_report
from animadao.version_checker import VersionChecker


def _merge_ignore(base: set[str] | None, extra: Iterable[str] | None) -> set[str]:
    out = set(s.lower() for s in (base or set()))
    out |= {s.lower() for s in (extra or [])}
    return out


@click.group(help="AnimaDao — dependency health checker.")
def cli() -> None: ...


@cli.command("scan")
@click.option(
    "--project",
    type=click.Path(path_type=Path, exists=True, file_okay=False),
    default=Path("."),
    help="Project root.",
)
@click.option(
    "--src",
    type=click.Path(path_type=Path, exists=True),
    default=None,
    help="Source root to scan imports.",
)
def scan_cmd(project: Path, src: Path | None) -> None:
    deps: list[Requirement] = load_declared_deps_any(project).requirements
    imports = find_top_level_imports(src or project)
    click.echo(
        json.dumps(
            {
                "declared": [r.name + (str(r.specifier) if str(r.specifier) else "") for r in deps],
                "imports": sorted(imports),
            },
            indent=2,
        )
    )


@cli.command("check")
@click.option(
    "--project",
    type=click.Path(path_type=Path, exists=True, file_okay=False),
    default=Path("."),
    help="Project root.",
)
@click.option(
    "--mode",
    type=click.Choice(["declared", "installed"]),
    default=None,
    help="What to compare against PyPI.",
)
@click.option("--ignore", multiple=True, help="Ignore packages (can repeat).")
@click.option("--pypi-ttl", type=int, default=None, help="PyPI cache TTL seconds (default 86400).")
@click.option("--pypi-concurrency", type=int, default=None, help="Parallel HTTP requests to PyPI (default 8).")
def check_cmd(
    project: Path,
    mode: str | None,
    ignore: tuple[str, ...],
    pypi_ttl: int | None,
    pypi_concurrency: int | None,
) -> None:
    cfg = load_config(project).with_overrides(mode=mode, ignore=ignore, ttl=pypi_ttl, conc=pypi_concurrency)

    checker = VersionChecker(ttl_seconds=cfg.pypi_ttl_seconds, concurrency=cfg.pypi_concurrency)
    if cfg.mode == "declared":
        declared = load_declared_deps_any(project).requirements
        outdated, unpinned = checker.check_declared(declared)
    else:
        from importlib import metadata as im

        installed = {d.metadata["Name"]: d.version for d in im.distributions()}
        outdated, unpinned = checker.check_installed(installed)

    ig = cfg.ignore_distributions or set()
    out = {
        "outdated": [o.__dict__ for o in outdated if o.name.lower() not in ig],
        "unpinned": [u.__dict__ for u in unpinned if u.name.lower() not in ig],
        "mode": cfg.mode,
    }
    click.echo(json.dumps(out, indent=2))


@cli.command("unused")
@click.option(
    "--project",
    type=click.Path(path_type=Path, exists=True, file_okay=False),
    default=Path("."),
    help="Project root.",
)
@click.option(
    "--src",
    type=click.Path(path_type=Path, exists=True),
    default=None,
    help="Source root to scan imports.",
)
@click.option("--ignore", multiple=True, help="Ignore packages (can repeat).")
def unused_cmd(project: Path, src: Path | None, ignore: tuple[str, ...]) -> None:
    declared = load_declared_deps_any(project).requirements
    imports = find_top_level_imports(src or project)
    ig = {s.lower() for s in ignore}
    unused = [u for u in guess_unused(declared, imports) if u.lower() not in ig]
    click.echo(json.dumps({"unused": unused}, indent=2))


@cli.command("report")
@click.option(
    "--project",
    type=click.Path(path_type=Path, exists=True, file_okay=False),
    default=Path("."),
    help="Project root.",
)
@click.option(
    "--src",
    type=click.Path(path_type=Path, exists=True),
    default=None,
    help="Source root to scan imports.",
)
@click.option("--out", type=click.Path(path_type=Path), default=None, help="Path to write report.")
@click.option("--mode", type=click.Choice(["declared", "installed"]), default=None, help="Report mode.")
@click.option("--ignore", multiple=True, help="Ignore packages (can repeat).")
@click.option(
    "--format",
    "fmt",
    type=click.Choice(["json", "md", "html"]),
    default="json",
    help="Output format.",
)
@click.option("--pypi-ttl", type=int, default=None, help="PyPI cache TTL seconds (default 86400).")
@click.option("--pypi-concurrency", type=int, default=None, help="Parallel HTTP requests to PyPI (default 8).")
def report_cmd(
    project: Path,
    src: Path | None,
    out: Path | None,
    mode: str | None,
    ignore: tuple[str, ...],
    fmt: str,
    pypi_ttl: int | None,
    pypi_concurrency: int | None,
) -> None:
    cfg = load_config(project).with_overrides(
        mode=mode,
        src=[str(src)] if src else None,
        ignore=ignore,
        ttl=pypi_ttl,
        conc=pypi_concurrency,
    )
    try:
        path = generate_report(
            project_root=project,
            src_root=src,
            out_path=out,
            mode=cfg.mode,
            ignore=cfg.ignore_distributions or set(),
            ttl_seconds=cfg.pypi_ttl_seconds,
            concurrency=cfg.pypi_concurrency,
            output_format=fmt,
        )
        click.echo(str(path))
    except Exception as exc:
        click.echo(f"ERROR: {exc}", err=True)
        sys.exit(1)



from __future__ import annotations

import json
from collections.abc import Iterable
from dataclasses import asdict
from pathlib import Path

from packaging.requirements import Requirement

from animadao.dependency_checker import guess_unused, load_declared_deps_any
from animadao.import_scanner import find_top_level_imports
from animadao.version_checker import VersionChecker


def _apply_ignore(names: Iterable[str], ignore: set[str] | None) -> list[str]:
    if not ignore:
        return list(names)
    ig = {s.lower() for s in ignore}
    return [n for n in names if n.lower() not in ig]


def _render_md(data: dict) -> str:
    lines: list[str] = []
    s = data["summary"]
    lines.append("# AnimaDao report\n")
    summary = (
        f"**declared:** {s['declared']}  \n"
        f"**imports:** {s['imports_found']}  \n"
        f"**outdated:** {s['outdated']}  \n"
        f"**unpinned:** {s['unpinned']}  \n"
        f"**unused:** {s['unused']}\n"
    )
    lines.append(summary)
    if data["outdated"]:
        lines.append("## Outdated\n\n| package | current | latest |\n|---|---:|---:|")
        for o in data["outdated"]:
            lines.append(f"| {o['name']} | {o['current']} | {o['latest']} |")
        lines.append("")
    if data["unpinned"]:
        lines.append("## Unpinned\n\n| package | spec |\n|---|---|")
        for u in data["unpinned"]:
            lines.append(f"| {u['name']} | `{u['spec']}` |")
        lines.append("")
    if data["unused"]:
        lines.append("## Unused\n\n" + ", ".join(sorted(data["unused"])) + "\n")
    return "\n".join(lines)


def _render_html(data: dict) -> str:
    def table(rows: list[list[str]], head: list[str]) -> str:
        th = "".join(f"<th>{h}</th>" for h in head)
        trs = "".join("<tr>" + "".join(f"<td>{c}</td>" for c in r) + "</tr>" for r in rows)
        return f"<table><thead><tr>{th}</tr></thead><tbody>{trs}</tbody></table>"

    s = data["summary"]
    parts = [
        "<!DOCTYPE html><html><head><meta charset='utf-8'><title>AnimaDao report</title>",
        (
            "<style>"
            "body{font-family:system-ui,monospace;padding:20px}\n"
            "table{border-collapse:collapse}\n"
            "td,th{border:1px solid #ddd;padding:6px 10px}"
            "</style>"
        ),
        "</head><body>",
        "<h1>AnimaDao report</h1>",
        f"<p><b>declared:</b> {s['declared']} &nbsp; <b>imports:</b> {s['imports_found']} "
        f"&nbsp; <b>outdated:</b> {s['outdated']} &nbsp; <b>unpinned:</b> {s['unpinned']} "
        f"&nbsp; <b>unused:</b> {s['unused']}</p>",
    ]
    if data["outdated"]:
        rows = [[o["name"], o["current"], o["latest"]] for o in data["outdated"]]
        parts += ["<h2>Outdated</h2>", table(rows, ["package", "current", "latest"])]
    if data["unpinned"]:
        rows = [[u["name"], u["spec"]] for u in data["unpinned"]]
        parts += ["<h2>Unpinned</h2>", table(rows, ["package", "spec"])]
    if data["unused"]:
        parts += ["<h2>Unused</h2>", "<p>" + ", ".join(sorted(data["unused"])) + "</p>"]
    parts.append("</body></html>")
    return "\n".join(parts)


def generate_report(
    project_root: Path,
    src_root: Path | None = None,
    out_path: Path | None = None,
    *,
    mode: str = "declared",  # declared | installed
    ignore: set[str] | None = None,  # игнор пакетов по имени (case-insensitive)
    ttl_seconds: int = 86400,
    concurrency: int = 8,
    output_format: str = "json",  # json | md | html
) -> Path:
    """
    Генерирует отчёт (json/md/html) по выбранному режиму.
    """
    if not (project_root / "pyproject.toml").is_file() and not (project_root / "requirements.txt").is_file():
        raise FileNotFoundError(f"No pyproject.toml or requirements.txt in: {project_root}")

    ignore = {s.lower() for s in (ignore or set())}

    declared_reqs: list[Requirement] = []
    imports: set[str] = set()

    if mode == "declared":
        declared_reqs = load_declared_deps_any(project_root).requirements
        imports = find_top_level_imports(src_root or project_root)
        # Version check on declared
        checker = VersionChecker(ttl_seconds=ttl_seconds, concurrency=concurrency)
        outdated, unpinned = checker.check_declared(declared_reqs)
    elif mode == "installed":
        # соберём установленные пакеты
        from importlib import metadata as im

        installed = {d.metadata["Name"]: d.version for d in im.distributions()}
        checker = VersionChecker(ttl_seconds=ttl_seconds, concurrency=concurrency)
        outdated, unpinned = checker.check_installed(installed)
        # для installed импорт-скан имеет меньший смысл, но оставим для консистентности
        imports = find_top_level_imports(src_root or project_root)
    else:
        raise ValueError("mode must be 'declared' or 'installed'")

    # Игнорируем пакеты в отчёте
    outdated = [o for o in outdated if o.name.lower() not in ignore]
    unpinned = [u for u in unpinned if u.name.lower() not in ignore]

    unused: list[str] = []
    if mode == "declared":
        unused = [u for u in guess_unused(declared_reqs, imports) if u.lower() not in ignore]

    data = {
        "summary": {
            "declared": len(declared_reqs) if mode == "declared" else 0,
            "imports_found": len(imports),
            "outdated": len(outdated),
            "unpinned": len(unpinned),
            "unused": len(unused),
        },
        "outdated": [asdict(o) for o in outdated],
        "unpinned": [asdict(u) for u in unpinned],
        "unused": unused,
        "imports": sorted(imports),
        "mode": mode,
    }

    # Пишем файл
    out = out_path or (project_root / ("report." + ("json" if output_format == "json" else output_format)))
    if output_format == "json":
        out.write_text(json.dumps(data, indent=2, ensure_ascii=False), encoding="utf-8")
    elif output_format == "md":
        out.write_text(_render_md(data), encoding="utf-8")
    elif output_format == "html":
        out.write_text(_render_html(data), encoding="utf-8")
    else:
        raise ValueError("output_format must be one of: json|md|html")
    return out



from __future__ import annotations

import ast
from pathlib import Path


def find_top_level_imports(src_root: Path) -> set[str]:
    """
    Walk Python files under `src_root` and collect top-level import names.
    We record only the top module part: e.g. `requests.adapters` -> `requests`.

    Returns:
        set[str]: unique top-level import names found.
    """
    imports: set[str] = set()
    for py in src_root.rglob("*.py"):
        try:
            text = py.read_text(encoding="utf-8")
        except Exception:
            continue
        try:
            tree = ast.parse(text, filename=str(py))
        except SyntaxError:
            continue

        for node in ast.walk(tree):
            if isinstance(node, ast.Import):
                for alias in node.names:
                    top = alias.name.split(".")[0]
                    if top:
                        imports.add(top)
            elif isinstance(node, ast.ImportFrom) and node.module:
                top = node.module.split(".")[0]
                if top:
                    imports.add(top)
    return imports



from __future__ import annotations

from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path

try:
    import tomllib as tomli  # Python 3.11+
except Exception:
    import tomli  # type: ignore

from packaging.requirements import Requirement


@dataclass(frozen=True)
class DeclaredDeps:
    """Container for declared dependencies."""

    requirements: list[Requirement]


def _normalize_dist_name(name: str) -> str:
    return name.lower().replace("-", "_")


# ---------- PEP 621 (project) ----------


def load_declared_deps(pyproject_path: Path) -> DeclaredDeps:
    """Existing PEP 621 loader (kept for backward compatibility)."""
    data = tomli.loads(pyproject_path.read_text(encoding="utf-8"))
    project = data.get("project", {}) or {}
    deps_raw: list[str] = list(project.get("dependencies") or [])
    opt = project.get("optional-dependencies") or {}
    for _extra, items in (opt or {}).items():
        deps_raw.extend(items or [])

    requirements: list[Requirement] = []
    for s in deps_raw:
        try:
            requirements.append(Requirement(s))
        except Exception:
            continue
    return DeclaredDeps(requirements=requirements)


# ---------- Poetry ([tool.poetry.dependencies]) ----------


def _caret_to_spec(ver: str) -> str:
    """
    Convert caret constraints ^X.Y.Z to PEP 440 range '>=X.Y.Z,<next' (basic).
    """
    import re

    parts = [int(p) for p in re.split(r"[._-]", ver) if p.isdigit()]
    while len(parts) < 3:
        parts.append(0)
    major, minor, patch = parts[:3]
    if major > 0:
        upper = f"{major + 1}.0.0"
    elif minor > 0:
        upper = f"0.{minor + 1}.0"
    else:
        upper = "0.1.0"
    return f">={major}.{minor}.{patch},<{upper}"


def _poetry_value_to_req(name: str, val: object) -> str:
    """
    Turn Poetry dep value into a Requirement string.
    - "2.31.0" -> "name==2.31.0"
    - "^1.26"  -> "name>=1.26,<2.0" (or next minor for 0.x)
    - "*" or {} -> "name"
    - {version="^1.2", extras=[...]} -> include just version (extras ignored for MVP)
    """
    if isinstance(val, str):
        if val == "*" or val.strip() == "":
            return name
        if val.startswith("^"):
            return f"{name}{_prepend_spec(_caret_to_spec(val[1:]))}"
        # raw exact version
        if val[0].isdigit():
            return f"{name}=={val}"
        # already a spec like ">=1.0"
        return f"{name}{_prepend_spec(val)}"
    if isinstance(val, dict):
        ver = str(val.get("version", "")).strip()
        if not ver or ver == "*":
            return name
        if ver.startswith("^"):
            return f"{name}{_prepend_spec(_caret_to_spec(ver[1:]))}"
        if ver[0].isdigit():
            return f"{name}=={ver}"
        return f"{name}{_prepend_spec(ver)}"
    return name


def _prepend_spec(spec: str) -> str:
    spec = spec.strip()
    return f"{spec}" if spec.startswith((">", "<", "=", "!")) else f"=={spec}"


def load_poetry_deps(pyproject_path: Path) -> DeclaredDeps:
    data = tomli.loads(pyproject_path.read_text(encoding="utf-8"))
    tool = data.get("tool", {}) or {}
    poetry = tool.get("poetry", {}) or {}
    deps = poetry.get("dependencies") or {}
    requirements: list[Requirement] = []
    for name, val in deps.items():
        if name.lower() == "python":
            continue
        try:
            req_str = _poetry_value_to_req(name, val)
            requirements.append(Requirement(req_str))
        except Exception:
            continue
    return DeclaredDeps(requirements=requirements)


# ---------- requirements.txt ----------


def _parse_requirements_file(path: Path, seen: set[Path] | None = None, depth: int = 0) -> list[str]:
    if seen is None:
        seen = set()
    if depth > 3:
        return []
    try:
        text = path.read_text(encoding="utf-8")
    except Exception:
        return []
    seen.add(path.resolve())
    out: list[str] = []
    for raw in text.splitlines():
        line = raw.strip()
        if not line or line.startswith("#"):
            continue
        if line.startswith("-r ") or line.startswith("--requirement "):
            ref = line.split(maxsplit=1)[1].strip()
            sub = (path.parent / ref).resolve()
            if sub not in seen:
                out.extend(_parse_requirements_file(sub, seen, depth + 1))
            continue
        out.append(line)
    return out


def load_requirements_txt(project_root: Path) -> DeclaredDeps:
    req_file = project_root / "requirements.txt"
    specs = _parse_requirements_file(req_file)
    requirements: list[Requirement] = []
    for s in specs:
        try:
            requirements.append(Requirement(s))
        except Exception:
            continue
    return DeclaredDeps(requirements=requirements)


# ---------- Универсальный лоадер ----------


def load_declared_deps_any(project_root: Path) -> DeclaredDeps:
    """
    Priority:
      1) pyproject.toml [project] (PEP 621)
      2) pyproject.toml [tool.poetry.dependencies]
      3) requirements.txt
    """
    pyproject = project_root / "pyproject.toml"
    if pyproject.is_file():
        data = tomli.loads(pyproject.read_text(encoding="utf-8"))
        if "project" in data:
            return load_declared_deps(pyproject)
        if "tool" in data and "poetry" in (data.get("tool") or {}):
            return load_poetry_deps(pyproject)
    req = project_root / "requirements.txt"
    if req.is_file():
        return load_requirements_txt(project_root)
    raise FileNotFoundError("No dependencies source found (pyproject or requirements.txt)")


def guess_unused(requirements: Iterable[Requirement], imported: Iterable[str]) -> list[str]:
    imported_set = {name.lower() for name in imported}
    unused: list[str] = []
    for req in requirements:
        name = req.name
        norm = _normalize_dist_name(name)
        candidates = {norm}
        if name.lower() in {"beautifulsoup4", "bs4"}:
            candidates.update({"bs4", "beautifulsoup4"})
        if imported_set.isdisjoint(candidates):
            unused.append(name)
    return sorted(unused)



from __future__ import annotations

from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path

try:
    import tomllib as tomli  # 3.11+
except Exception:  # 3.10
    import tomli  # type: ignore


@dataclass(frozen=True)
class Config:
    mode: str = "declared"  # declared | installed
    src: list[str] = None  # список путей; если None -> [project_root]
    ignore_distributions: set[str] = None  # lower-case имена дистрибутивов
    ignore_imports: set[str] = None  # lower-case имена импортов
    pypi_ttl_seconds: int = 86400  # кеш PyPI (по умолчанию сутки)
    pypi_concurrency: int = 8  # параллелизм запросов к PyPI

    def with_overrides(
        self,
        mode: str | None = None,
        src: Iterable[str] | None = None,
        ignore: Iterable[str] | None = None,
        ttl: int | None = None,
        conc: int | None = None,
    ) -> Config:
        return Config(
            mode=mode or self.mode,
            src=list(src) if src is not None else (self.src or []),
            ignore_distributions=set(self.ignore_distributions or set())
            | ({s.lower() for s in ignore} if ignore else set()),
            ignore_imports=self.ignore_imports or set(),
            pypi_ttl_seconds=ttl if ttl is not None else self.pypi_ttl_seconds,
            pypi_concurrency=conc if conc is not None else self.pypi_concurrency,
        )


def _load_one(path: Path) -> dict:
    try:
        return tomli.loads(path.read_text(encoding="utf-8"))
    except Exception:
        return {}


def load_config(project_root: Path) -> Config:
    """
    Загружает конфиг в порядке приоритета:
    1) <project>/.animadao.toml
    2) $HOME/.config/animadao/config.toml
    Отсутствующие поля -> значения по умолчанию.
    """
    conf = Config()
    candidates = [
        project_root / ".animadao.toml",
        Path.home() / ".config" / "animadao" / "config.toml",
    ]
    data: dict = {}
    for p in candidates:
        if p.is_file():
            data = _load_one(p)
            break

    core = data.get("core", {}) or {}
    ignore = data.get("ignore", {}) or {}

    mode = str(core.get("mode", conf.mode))
    src = core.get("src")
    if isinstance(src, str):
        src = [src]
    src_list = [str(s) for s in (src or [])]

    ttl = int(core.get("pypi_ttl_seconds", conf.pypi_ttl_seconds))
    conc = int(core.get("pypi_concurrency", conf.pypi_concurrency))

    ig_dist = {s.lower() for s in (ignore.get("distributions") or [])}
    ig_imp = {s.lower() for s in (ignore.get("imports") or [])}

    return Config(
        mode=mode if mode in {"declared", "installed"} else "declared",
        src=src_list or None,
        ignore_distributions=ig_dist or None,
        ignore_imports=ig_imp or None,
        pypi_ttl_seconds=ttl,
        pypi_concurrency=conc,
    )
